use serde::{Serialize, Deserialize};
use std::{fs, path::Path};

#[derive(Clone, Serialize, Deserialize, Debug, PartialEq)]
pub struct Config {
    // Pulled from configuration file
    pub main: MainConfig,
    // Pulled from auto-created yaml files
    pub client_sessions: Vec<ClientSession>,
}

#[derive(Clone, Serialize, Deserialize, Debug, PartialEq)]
#[serde_with::skip_serializing_none]
pub struct MainConfig {
    pub bind_address: Option<String>,
    pub bind_port: Option<String>,
    pub client_storage_path: Option<String>,
    pub allowed_receivers: Option<Vec<String>>,
    pub clients: Vec<ClientConfig>,
    pub mappings: Vec<Mapping>,
}

#[derive(Clone, Serialize, Deserialize, Debug, PartialEq)]
pub struct ClientConfig {
    pub mxid: String,
    pub password: String,
}

#[derive(Clone, Serialize, Deserialize, Debug, PartialEq)]
pub struct ClientSession {
    pub mxid: String,
    pub device_id: String,
    pub access_token: String,
}

#[derive(Clone, Serialize, Deserialize, Debug, PartialEq)]
pub struct Mapping {
    pub from: Option<String>,
    pub to: Option<String>,
    pub mxid_sender: Option<String>,
    pub room: String,
}

pub fn load_config(file_path: &str) -> Config {
    let yaml_config = fs::read_to_string(file_path).expect("Error while reading config file");
    let mut cfg: MainConfig = serde_yaml::from_str(&yaml_config).expect("Config file not in expected format");

    // Set default values if necessary
    if cfg.bind_address.is_none() {
        cfg.bind_address = Some("127.0.0.1".to_string());
    }
    if cfg.bind_port.is_none() {
        cfg.bind_port = Some("25".to_string());
    }
    if cfg.client_storage_path.is_none() {
        let path = Path::new(file_path).parent().unwrap().join("client_storage");
        cfg.client_storage_path = Some(path.to_str().unwrap().to_string());
    }

    let sessions: Vec<ClientSession> = match fs::read_to_string(&format!("{}.sessions/sessions.yaml", file_path)) {
        Ok(string) => serde_yaml::from_str(&string).expect("Sessions file not in expected format"),
        Err(e) => vec![],
    };

    Config {main: cfg, client_sessions: sessions}
}

pub fn write_client_sessions(new: &Vec<ClientSession>, old: &Vec<ClientSession>, config_file_path: &str) {
    let new_yaml_str = serde_yaml::to_string(new).expect("Failed to serialize new config.");
    let old_yaml_str = serde_yaml::to_string(old).expect("Failed to serialize old config.");
    if new_yaml_str != old_yaml_str {
        rotate_client_sessions(config_file_path);
        fs::write(&format!("{}.sessions/sessions.yaml", config_file_path), &new_yaml_str).expect("Error writing new config to file");
    }
}

fn rotate_client_sessions(base_path: &str) {
    match std::fs::create_dir(format!("{}.sessions", base_path)) {
        Ok(_) => (),
        Err(_e) => (),
    };
    for i in 9..2 {
        std::fs::copy(&format!("{}.sessions/sessions.yaml.{}", base_path, i - 1), &format!("{}.{}", base_path, i)).expect("");
    }
    std::fs::copy(base_path, &format!("{}.1", base_path)).expect("");
}